"""Patch generators for edges."""


from itertools import product
from typing import Dict, Iterable, List

import numpy as np
import pandas as pd
from matplotlib.patches import Arc, Circle, Patch, Path, PathPatch

from nxviz.geometry import correct_hive_angles
from nxviz.polcart import to_cartesian, to_polar, to_radians


def circos(
    et: pd.DataFrame,
    pos: Dict,
    edge_color: Iterable,
    alpha: Iterable,
    lw: Iterable,
    aes_kw: Dict,
) -> List[Patch]:
    """Circos plot line drawing."""
    patches = []
    for r, d in et.iterrows():
        verts = [pos[d["source"]], (0, 0), pos[d["target"]]]
        codes = [Path.MOVETO, Path.CURVE3, Path.CURVE3]
        path = Path(verts, codes)
        patch = PathPatch(
            path, edgecolor=edge_color[r], alpha=alpha[r], lw=lw[r], **aes_kw
        )
        patches.append(patch)
    return patches


def line(
    et: pd.DataFrame,
    pos: Dict,
    edge_color: Iterable,
    alpha: Iterable,
    lw: Iterable,
    aes_kw: Dict,
):
    """Straight line drawing function."""
    patches = []
    for r, d in et.iterrows():
        start = d["source"]
        end = d["target"]
        verts = [pos[start], pos[end]]
        codes = [Path.MOVETO, Path.LINETO]
        path = Path(verts, codes)
        patch = PathPatch(
            path, edgecolor=edge_color[r], alpha=alpha[r], lw=lw[r], **aes_kw
        )
        patches.append(patch)
    return patches


def arc(
    et: pd.DataFrame,
    pos: Dict,
    edge_color: Iterable,
    alpha: Iterable,
    lw: Iterable,
    aes_kw: Dict,
):
    """Arc plot edge drawing function."""
    patches = []
    for r, d in et.iterrows():
        start = d["source"]
        end = d["target"]
        start_x, start_y = pos[start]
        end_x, end_y = pos[end]

        middle_x = np.mean([start_x, end_x])
        middle_y = np.mean([start_y, end_y])

        width = abs(end_x - start_x)
        height = width

        r1, theta1 = to_polar(start_x - middle_x, start_y - middle_y)
        r2, theta2 = to_polar(end_x - middle_x, end_y - middle_y)

        theta1 = np.rad2deg(theta1)
        theta2 = np.rad2deg(theta2)

        theta1, theta2 = min([theta1, theta2]), max([theta1, theta2])

        patch = Arc(
            xy=(middle_x, middle_y),
            width=width,
            height=height,
            theta1=theta1,
            theta2=theta2,
            edgecolor=edge_color[r],
            alpha=alpha[r],
            lw=lw[r],
            **aes_kw,
        )
        patches.append(patch)
    return patches


def hive(
    et: pd.DataFrame,
    pos: Dict,
    pos_cloned: Dict,
    edge_color: Iterable,
    alpha: Iterable,
    lw: Iterable,
    aes_kw: Dict,
    curves: bool = True,
):
    """Hive plot line drawing function."""
    rad = pd.Series(pos).apply(lambda val: to_polar(*val)).to_dict()
    if pos_cloned is None:
        pos_cloned = pos
    rad_cloned = pd.Series(pos_cloned).apply(lambda val: to_polar(*val)).to_dict()

    patches = []
    for r, d in et.iterrows():
        start = d["source"]
        end = d["target"]
        start_radius, start_theta = rad[start]
        end_radius, end_theta = rad[end]

        _, start_theta_cloned = rad_cloned[start]
        _, end_theta_cloned = rad_cloned[end]

        # Find the pair of start and end thetas that give the smallest acute angle
        smallest_pair = None
        smallest_nonzero_angle = np.inf
        starts = [start_theta, start_theta_cloned]
        ends = [end_theta, end_theta_cloned]

        for start, end in product(starts, ends):
            start, end = correct_hive_angles(start, end)
            if not np.allclose(end - start, 0):
                angle = to_radians(abs(min([end - start, start - end])))
                if angle < smallest_nonzero_angle:
                    smallest_nonzero_angle = abs(angle)
                    smallest_pair = ((start_radius, start), (end_radius, end))

        if smallest_pair is None:
            continue

        (start_radius, start_theta), (end_radius, end_theta) = smallest_pair
        if np.allclose(end_theta, 0):
            end_theta = 2 * np.pi
        startx, starty = to_cartesian(start_radius, start_theta)
        endx, endy = to_cartesian(end_radius, end_theta)

        middle_theta = np.mean([start_theta, end_theta])

        middlex1, middley1 = to_cartesian(start_radius, middle_theta)
        middlex2, middley2 = to_cartesian(end_radius, middle_theta)
        endx, endy = to_cartesian(end_radius, end_theta)

        verts = [(startx, starty), (endx, endy)]
        codes = [Path.MOVETO, Path.LINETO]
        if curves:
            verts = [
                (startx, starty),
                (middlex1, middley1),
                (middlex2, middley2),
                (endx, endy),
            ]
            codes = [
                Path.MOVETO,
                Path.CURVE4,
                Path.CURVE4,
                Path.CURVE4,
            ]

        path = Path(verts, codes)
        patch = PathPatch(
            path, lw=lw[r], alpha=alpha[r], edgecolor=edge_color[r], **aes_kw
        )
        patches.append(patch)
    return patches


def matrix(
    et,
    pos,
    pos_cloned,
    edge_color: Iterable,
    alpha: Iterable,
    lw: Iterable,
    aes_kw: Dict,
):
    """Matrix plot edge drawing function."""
    patches = []
    for r, d in et.iterrows():
        start = d["source"]
        end = d["target"]
        x_start, y_start = pos_cloned[start]
        x_end, y_end = pos[end]

        x, y = (max(x_start, y_start), max(x_end, y_end))
        kw = {
            "fc": edge_color[r],
            "alpha": alpha[r],
            "radius": lw[r],
            "zorder": 1,
        }
        kw.update(aes_kw)
        patch = Circle(xy=(x, y), **kw)
        patches.append(patch)
    return patches



"""Declarative graph visualization facets.

There are two groups of functions in here:
One are intended to yield graph objects that
contain the subset of nodes and edges to be plotted.
The others are faceting functions that can be called on.
The faceting functions rely on the high level API for plotting.
"""
import warnings
from functools import partial, update_wrapper
from itertools import combinations
from typing import Callable, Hashable

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

from nxviz import annotate, api, utils


### Iterators to generate subgraphs.
def hive_triplets(G: nx.Graph, group_by: Hashable):
    """Yield subgraphs containing triplets of node categories.

    Intended for hive plotting.
    """
    nt = utils.node_table(G)
    groups = sorted(nt[group_by].unique())
    if len(groups) > 6:
        warnings.warn(
            "You have more than 6 groups of nodes, "
            "which means you might end up having a lot of subplots made. "
            "User beware! "
            "We recommend using hive plots only when you have 6 or fewer "
            "groups of nodes."
        )
    triplets = combinations(groups, 3)

    for groups in triplets:
        wanted_nodes = (n for n in G.nodes() if G.nodes[n][group_by] in groups)
        yield G.subgraph(wanted_nodes), groups


def edge_group(G: nx.Graph, group_by: Hashable):
    """Yield graphs containing only certain categories of edges."""
    et = utils.edge_table(G)
    groups = sorted(et[group_by].unique())
    for group in groups:
        G_sub = G.copy()
        G_sub.remove_edges_from(G_sub.edges())
        for u, v, d in G.edges(data=True):
            if d[group_by] == group:
                G_sub.add_edge(u, v, **d)
        yield G_sub, group


def node_group_edges(G: nx.Graph, group_by: Hashable):
    """Return a subgraph containing edges connected to a particular category of nodes."""
    nt = utils.node_table(G)
    groups = sorted(nt[group_by].unique())

    for group in groups:
        G_sub = G.copy()
        G_sub.remove_edges_from(G_sub.edges())

        wanted_nodes = (n for n in G.nodes() if G.nodes[n][group_by] == group)
        for node in wanted_nodes:
            for u, v, d in G.edges(node, data=True):
                G_sub.add_edge(u, v, **d)
        yield G_sub, group


def n_rows_cols(groups):
    """Return squarest n_rows and n_cols combination."""
    nrows = ncols = int(np.ceil(np.sqrt(len(groups))))
    return nrows, ncols


def null(*args, **kwargs):
    """A passthrough function that does nothing."""
    pass


grouping_annotations = {
    "api.circos": annotate.circos_group,
    "api.arc": annotate.arc_group,
    "api.matrix": annotate.matrix_block,
    "api.hive": annotate.hive_group,
    "api.parallel": annotate.parallel_group,
}

node_color_annotations = {
    "api.hive": null,
    "api.circos": annotate.node_colormapping,
    "api.matrix": annotate.node_colormapping,
    "api.arc": annotate.node_colormapping,
    "api.parallel": annotate.node_colormapping,
}


def facet_plot(
    G: nx.Graph,
    plotting_func: Callable,
    node_facet_func: Callable,
    node_group_by: Hashable,
    node_sort_by: Hashable,
    node_color_by: Hashable,
    edge_facet_func: Callable,
    edge_group_by: Hashable,
    edge_color_by: Hashable,
):
    """Generic facet plotting function.

    All faceting funcs should take G and group_by and yield graphs.
    Underneath the hood, how they work shouldn't be of concern.

    Edge facet func takes priority if both node and edge facet func are specified.
    Just keep this in mind.

    ## Parameters

    - `G`: The graph to facet.
    - `plotting_func`: One of the high level API functions to use for plotting.
    - `node_facet_func`: A function to facet the nodes by.
    - `node_group_by`: Node metadata attribute to group nodes by.
    - `node_sort_by`: Node metadata attribute to sort nodes by.
    - `node_color_by`: Node metadata attribute to color nodes by.
    - `edge_facet_func`: A function to facet the edges by.
    - `edge_group_by`: Edge metadata attribute to group edges by.
    - `edge_color_by`: Edge metadata attribute to color edges by.
    """
    group_by = node_group_by
    facet_func = node_facet_func
    if edge_facet_func and edge_group_by:
        group_by = edge_group_by
        facet_func = edge_facet_func

    graphs, groups = zip(*facet_func(G, group_by))
    nrows, ncols = n_rows_cols(groups)
    fig, axes = plt.subplots(figsize=(3 * nrows, 3 * ncols), nrows=nrows, ncols=ncols)
    axes = list(axes.flatten())

    for G_sub, group, ax in zip(graphs, groups, axes):
        plt.sca(ax)
        plotting_func(
            G_sub,
            group_by=node_group_by,
            sort_by=node_sort_by,
            node_color_by=node_color_by,
            edge_color_by=edge_color_by,
        )
        if node_group_by:
            grouping_annotations.get(plotting_func.__name__)(
                G_sub, group_by=node_group_by
            )
        if node_color_by:
            # Annotate only on the left most axes
            idx = axes.index(ax)
            if not idx % nrows:
                node_color_annotations[plotting_func.__name__](G, node_color_by)
        ax.set_title(f"{group_by} = {group}")

    last_idx = axes.index(ax)
    for ax in axes[last_idx + 1 :]:
        fig.delaxes(ax)
    plt.tight_layout()


#### Function begins below. This belongs to the "edge faceting" category.

hive_panel = partial(
    facet_plot,
    plotting_func=api.hive,
    node_facet_func=hive_triplets,
    node_sort_by=None,
    node_color_by=None,
    edge_facet_func=None,
    edge_group_by=None,
    edge_color_by=None,
)
update_wrapper(hive_panel, facet_plot)
hive_panel.__name__ = "facet.hive_panel"

matrix_panel = partial(
    facet_plot,
    plotting_func=api.matrix,
    node_facet_func=None,
    node_group_by=None,
    node_sort_by=None,
    node_color_by=None,
    edge_facet_func=edge_group,
    edge_color_by=None,
)
update_wrapper(matrix_panel, facet_plot)
matrix_panel.__name__ = "facet.matrix_panel"

arc_panel = partial(
    facet_plot,
    plotting_func=api.arc,
    node_facet_func=None,
    node_group_by=None,
    node_sort_by=None,
    node_color_by=None,
    edge_facet_func=edge_group,
    edge_color_by=None,
)
update_wrapper(arc_panel, facet_plot)
arc_panel.__name__ = "facet.arc_panel"

circos_panel = partial(
    facet_plot,
    plotting_func=api.circos,
    node_facet_func=None,
    node_group_by=None,
    node_sort_by=None,
    node_color_by=None,
    edge_facet_func=edge_group,
    edge_color_by=None,
)
update_wrapper(circos_panel, facet_plot)
circos_panel.__name__ = "facet.circos_panel"

parallel_panel = partial(
    facet_plot,
    plotting_func=api.parallel,
    node_facet_func=None,
    node_group_by=None,
    node_sort_by=None,
    node_color_by=None,
    edge_facet_func=edge_group,
    edge_color_by=None,
)
update_wrapper(parallel_panel, facet_plot)
parallel_panel.__name__ = "facet.parallel_panel"



"""Node layout engines.

Compatible with NetworkX's drawing facilities;
they will each return a dictionary
of node mapped to a cartesian coordinate on the (x, y) plane.
"""

from typing import Dict, Hashable

import numpy as np
import pandas as pd

from nxviz.geometry import circos_radius, item_theta
from nxviz.polcart import to_cartesian
from nxviz.utils import group_and_sort


def parallel(
    nt: pd.DataFrame, group_by: Hashable, sort_by: Hashable = None
) -> Dict[Hashable, np.ndarray]:
    """Parallel coordinates node layout."""
    pos = dict()

    for x, (grp, data) in enumerate(nt.groupby(group_by)):
        if sort_by is not None:
            data = data.sort_values(sort_by)
        for y, (node, d) in enumerate(data.iterrows()):
            pos[node] = np.array([x * 4, y])
    return pos


def circos(
    nt: pd.DataFrame,
    group_by: Hashable = None,
    sort_by: Hashable = None,
    radius: float = None,
) -> Dict[Hashable, np.ndarray]:
    """Circos plot node layout."""
    pos = dict()
    nt = group_and_sort(nt, group_by, sort_by)
    nodes = list(nt.index)
    if radius is None:
        radius = circos_radius(len(nodes))
    if group_by:
        for grp, df in nt.groupby(group_by):
            for node, data in df.iterrows():
                x, y = to_cartesian(r=radius, theta=item_theta(nodes, node))
                pos[node] = np.array([x, y])
    else:
        for node, data in nt.iterrows():
            x, y = to_cartesian(r=radius, theta=item_theta(nodes, node))
            pos[node] = np.array([x, y])
    return pos


def hive(
    nt: pd.DataFrame,
    group_by,
    sort_by: Hashable = None,
    inner_radius: float = 8,
    rotation: float = 0,
):
    """Hive plot node layout.

    ## Parameters

    - `inner_radius`: The inner
    """
    nt = group_and_sort(nt, group_by=group_by, sort_by=sort_by)
    groups = sorted(nt[group_by].unique())
    if len(groups) > 3:
        raise ValueError(
            f"group_by {group_by} is associated with more than 3 groups. "
            f"The groups are {groups}. "
            "Hive plots can only handle at most 3 groups at a time."
        )
    pos = dict()
    for grp, df in nt.groupby(group_by):
        for i, (node, data) in enumerate(df.iterrows()):
            radius = inner_radius + i
            theta = item_theta(groups, grp) + rotation
            x, y = to_cartesian(r=radius * 2, theta=theta)
            pos[node] = np.array([x, y])
    return pos


def arc(nt, group_by: Hashable = None, sort_by: Hashable = None):
    """Arc plot node layout."""
    nt = group_and_sort(nt, group_by=group_by, sort_by=sort_by)
    pos = dict()
    for x, (node, data) in enumerate(nt.iterrows()):
        pos[node] = np.array([x * 2, 0])
    return pos


def geo(nt, group_by=None, sort_by=None, longitude="longitude", latitude="latitude"):
    """Geographical node layout."""
    pos = dict()
    for node, data in nt.iterrows():
        pos[node] = data[[longitude, latitude]]
    return pos


def matrix(nt, group_by: Hashable = None, sort_by: Hashable = None, axis="x"):
    """Matrix plot layout."""
    # Nodes should be grouped and sorted before we begin assigning coordinates.
    nt = group_and_sort(node_table=nt, group_by=group_by, sort_by=sort_by)

    # We are eventually going to return this pos dictionary.
    pos = dict()

    # Loop over each of the rows, and assign x, y coordinates in order of them being grouped and sorted.
    for i, (node, data) in enumerate(nt.iterrows()):
        x = (i + 1) * 2
        y = 0

        if axis == "y":
            x, y = y, x
        pos[node] = np.array([x, y])
    return pos



"""Node drawing functions."""

from copy import deepcopy
from functools import partial, update_wrapper
from typing import Callable, Dict, Hashable, Optional, Tuple

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
from matplotlib.patches import Circle

from nxviz import encodings, layouts
from nxviz.utils import node_table
from nxviz.plots import rescale, rescale_arc, rescale_square


def node_colors(nt: pd.DataFrame, color_by: Hashable):
    """Return pandas Series of node colors."""
    if color_by:
        return encodings.data_color(nt[color_by], nt[color_by])
    return pd.Series(["blue"] * len(nt), name="color_by", index=nt.index)


def transparency(
    nt: pd.DataFrame, alpha_by: Hashable, alpha_bounds: Optional[Tuple] = None
):
    """Return pandas Series of transparency (alpha) values.

    Transparency must always be normalized to (0, 1)."""
    if alpha_by is not None:
        ref_data = nt[alpha_by]
        if isinstance(alpha_bounds, tuple):
            ref_data = pd.Series(alpha_bounds)

        return encodings.data_transparency(nt[alpha_by], ref_data)
    return pd.Series([1.0] * len(nt), name="transparency", index=nt.index)


def node_size(nt: pd.DataFrame, size_by: Hashable):
    """Return pandas Series of node sizes."""
    if size_by:
        return encodings.data_size(nt[size_by], nt[size_by])
    return pd.Series([1.0] * len(nt), name="size", index=nt.index)


def node_glyphs(nt, pos, node_color, alpha, size, **encodings_kwargs):
    """Draw circos glyphs to the matplotlib axes object."""
    patches = dict()
    for r, d in nt.iterrows():
        kw = {"fc": node_color[r], "alpha": alpha[r], "radius": size[r], "zorder": 2}
        kw.update(encodings_kwargs)
        c = Circle(xy=pos[r], **kw)
        patches[r] = c
    return pd.Series(patches)


def draw(
    G: nx.Graph,
    layout_func: Callable,
    group_by: Hashable,
    sort_by: Hashable,
    color_by: Hashable = None,
    alpha_by: Hashable = None,
    size_by: Hashable = None,
    layout_kwargs: Dict = {},
    encodings_kwargs: Dict = {},
    rescale_func=rescale,
    ax=None,
):
    """Draw nodes to matplotlib axes.

    ## Parameters

    - `G`: The graph to plot.
    - `layout_func`: One of the node layout functions from `nxviz.layout`.
    - `group_by`: Categorical attribute key to group nodes by.
    - `sort_by`: Quantitative or ordinal attribute key to sort nodes.
    - `color_by`: Node attribute key to color nodes by.
    - `alpha_by`: Quantitative node attribute key to set transparency.
    - `size_by`: Quantitative node attribute key to set node size.
    - `layout_kwargs`: Keyword arguments to pass
        to the appropriate layout function.
    - `encodings_kwargs`: A dictionary of kwargs
        to determine the visual properties of the node.
    - `linefunc_kwargs`: All other keyword arguments passed in
        will be passed onto the appropriate linefunc.

    Special keyword arguments for `encodings_kwargs` include:

    - `size_scale`: A scaling factor for all node radii.
        Equivalent to multiplying all node radii by this number.
    - `alpha_scale`: A scaling factor for all nodes' transparencies.
        Equivalent to multiplying all alphas by this number.
        The default transparency is 1.0.
        If you need to make the nodes transparent,
        use a value smaller than one.
    - `alpha_bounds`: The bounds for transparency.
        Should be a tuple of `(lower, upper)` numbers.
        This keyword argument lets us manually set the bounds
        that we wish to have for 0 opacity (i.e. transparent)
        to 1.0 opacity (i.e. opaque.)

    Everything else passed in here will be passed
    to the matplotlib Patch constructor;
    see `nxviz.lines` for more information.
    """
    if ax is None:
        ax = plt.gca()
    nt = node_table(G)
    pos = layout_func(nt, group_by, sort_by, **layout_kwargs)
    node_color = node_colors(nt, color_by)

    encodings_kwargs = deepcopy(encodings_kwargs)
    alpha = transparency(nt, alpha_by) * encodings_kwargs.pop("alpha_scale", 1)
    size = node_size(nt, size_by) * encodings_kwargs.pop("size_scale", 1)
    patches = node_glyphs(nt, pos, node_color, alpha, size, **encodings_kwargs)
    for patch in patches:
        ax.add_patch(patch)

    rescale_func(G)
    return pos


hive = partial(
    draw,
    layout_func=layouts.hive,
    sort_by=None,
    layout_kwargs={"inner_radius": 8},
    encodings_kwargs={"size_scale": 0.5},
    rescale_func=rescale_square,
)
update_wrapper(hive, draw)
hive.__name__ = "nodes.hive"

circos = partial(
    draw,
    layout_func=layouts.circos,
    group_by=None,
    sort_by=None,
)
update_wrapper(circos, draw)
circos.__name__ = "nodes.circos"

arc = partial(
    draw,
    layout_func=layouts.arc,
    group_by=None,
    sort_by=None,
    rescale_func=rescale_arc,
)
update_wrapper(arc, draw)
arc.__name__ = "nodes.arc"

parallel = partial(
    draw,
    layout_func=layouts.parallel,
    sort_by=None,
    encodings_kwargs={"size_scale": 0.5},
)
update_wrapper(parallel, draw)
parallel.__name__ = "nodes.parallel"

matrix = partial(draw, layout_func=layouts.matrix, group_by=None, sort_by=None)
update_wrapper(matrix, draw)
matrix.__name__ = "nodes.matrix"

geo = partial(
    draw,
    layout_func=layouts.geo,
    group_by=None,
    sort_by=None,
    encodings_kwargs={"size_scale": 0.0015},
)
update_wrapper(geo, draw)
geo.__name__ = "nodes.geo"



"""
Utility geometry functions that can help with drawing to screen.
"""
import numpy as np

from .polcart import to_cartesian
from typing import List, Hashable


def item_theta(nodelist: List[Hashable], node: Hashable):
    """
    Maps node to an angle in radians.

    :param nodelist: Nodelist from the graph.
    :param node: The node of interest. Must be in the nodelist.
    :returns: theta -- the angle of the node in radians.
    """
    assert len(nodelist) > 0, "nodelist must be a list of items."
    assert node in nodelist, "node must be inside nodelist."

    i = nodelist.index(node)
    theta = i * 2 * np.pi / len(nodelist)

    return theta


def get_cartesian(r: float, theta: float):
    """
    Returns the cartesian (x,y) coordinates of (r, theta).

    :param r: Real-valued radius.
    :param theta: Angle in radians.
    :returns: to_cartesian(r, theta)
    """
    return to_cartesian(r, theta)


def correct_negative_angle(angle):
    """
    Corrects a negative angle to a positive one.

    :param angle: The angle in radians.
    :returns: `angle`, corrected to be positively-valued.
    """
    angle = angle % (2 * np.pi)
    if angle < 0:
        angle += 2 * np.pi
    return angle


def circos_radius(n_nodes: int, node_radius: float = 1.0):
    """
    Automatically computes the origin-to-node centre radius of the Circos plot
    using the triangle equality sine rule.

    a / sin(A) = b / sin(B) = c / sin(C)

    :param n_nodes: the number of nodes in the plot.
    :param node_r: the radius of each node.
    :returns: Origin-to-node centre radius.
    """
    A = 2 * np.pi / n_nodes  # noqa
    B = (np.pi - A) / 2  # noqa
    a = 2 * node_radius
    return a * np.sin(B) / np.sin(A)


def correct_hive_angles(start, end):
    """Perform correction of hive plot angles for edge drawing."""
    if start > np.pi and end == 0.0:
        end = 2 * np.pi
    if start < np.pi and end == 0.0:
        start, end = end, start
    if end < np.pi and start == 2 * np.pi:
        start = 0
    if end > np.pi and start == 0:
        start = 2 * np.pi
    return start, end



"""Functions for plotting."""

import matplotlib.pyplot as plt
import networkx as nx


def despine(ax=None):
    """Remove all spines (and ticks) from the matplotlib axes."""
    if ax is None:
        ax = plt.gca()
    for spine in ax.spines:
        ax.spines[spine].set_visible(False)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)


def respine(ax=None):
    """Reinstate the visibility of the spines.

    Use this when you want to make
    extremely fine-grained customizations to the plot axes
    and need to know where exactly to put markers.
    """
    if ax is None:
        ax = plt.gca()
    for spine in ax.spines:
        ax.spines[spine].set_visible(True)
    ax.xaxis.set_visible(True)
    ax.yaxis.set_visible(True)


def aspect_equal(ax=None):
    """Set aspect ratio of an axes object to be equal."""
    if ax is None:
        ax = plt.gca()
    ax.set_aspect("equal")


# The rescaling functions rescale the matplotlib axes.
# They all accept a graph, so that data-dependent xlim and ylims


def rescale(G: nx.Graph):
    """Default rescale."""
    ax = plt.gca()
    ax.relim()
    ax.autoscale_view()


def rescale_arc(G: nx.Graph):
    """Axes rescale function for arc plot."""
    ax = plt.gca()
    ax.relim()
    ymin, ymax = ax.get_ylim()
    maxheight = int(len(G)) + 1
    ax.set_ylim(ymin - 1, maxheight)
    ax.set_xlim(-1, len(G) * 2 + 1)


def rescale_square(G):
    """Axes rescale function to go square."""
    rescale(G)
    ax = plt.gca()
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    newmax = max([xmax, ymax, -xmin, -ymin])
    ax.set_xlim(-newmax, newmax)
    ax.set_ylim(-newmax, newmax)



"""Utility functions."""
from collections import Counter

import pandas as pd
import warnings
from typing import Iterable


def is_data_homogenous(data_container: Iterable):
    """
    Checks that all of the data in the container are of the same Python data
    type. This function is called in every other function below, and as such
    need not necessarily be called.

    :param data_container: A generic container of data points.
    """
    data_types = set([type(i) for i in data_container])
    return len(data_types) == 1


def infer_data_type(data_container: Iterable):
    """
    For a given container of data, infer the type of data as one of
    continuous, categorical, or ordinal.

    For now, it is a one-to-one mapping as such:

    - str:   categorical
    - int:   ordinal
    - float: continuous

    There may be better ways that are not currently implemented below. For
    example, with a list of numbers, we can check whether the number of unique
    entries is less than or equal to 12, but has over 10000+ entries. This
    would be a good candidate for floats being categorical.

    :param data_container: A generic container of data points.
    :type data_container: `iterable`

    """
    warnings.warn(
        "`infer_data_type` is deprecated! " "Please use `infer_data_family` instead!"
    )
    # Defensive programming checks.
    # 0. Ensure that we are dealing with lists or tuples, and nothing else.
    assert isinstance(data_container, list) or isinstance(
        data_container, tuple
    ), "data_container should be a list or tuple."
    # 1. Don't want to deal with only single values.
    assert (
        len(set(data_container)) > 1
    ), "There should be more than one value in the data container."
    # 2. Don't want to deal with mixed data.
    assert is_data_homogenous(data_container), "Data are not of a homogenous type!"

    # Once we check that the data type of the container is homogenous, we only
    # need to check the first element in the data container for its type.
    datum = data_container[0]

    # Return statements below
    # treat binomial data as categorical
    # TODO: make tests for this.
    if len(set(data_container)) == 2:
        return "categorical"

    elif isinstance(datum, str):
        return "categorical"

    elif isinstance(datum, int):
        return "ordinal"

    elif isinstance(datum, float):
        return "continuous"

    else:
        raise ValueError("Not possible to tell what the data type is.")


def infer_data_family(data: pd.Series):
    """Infer data family from a column of data.

    The three families are "continuous", "ordinal", and "categorical".

    The rules:

    - dtype = float:
        - min < 0 and max > 0: divergent
        - otherwise: continuous
    - dtype = integer:
        - greater than 12 distinct integers: continuous
        - otherwise: ordinal
    - dtype = object: categorical
    """
    if data.dtype == float:
        if data.min() < 0 and data.max() > 0:
            return "divergent"
        return "continuous"
    if data.dtype == int:
        if len(set(data)) > 9:
            return "continuous"
        return "ordinal"
    return "categorical"


def is_data_diverging(data_container: Iterable):
    """
    We want to use this to check whether the data are diverging or not.

    This is a simple check, can be made much more sophisticated.

    :param data_container: A generic container of data points.
    :type data_container: `iterable`
    """
    assert infer_data_type(data_container) in [
        "ordinal",
        "continuous",
    ], "Data type should be ordinal or continuous"

    # Check whether the data contains negative and positive values.
    has_negative = False
    has_positive = False
    for i in data_container:
        if i < 0:
            has_negative = True
        elif i > 0:
            has_positive = True
    if has_negative and has_positive:
        return True
    else:
        return False


def is_groupable(data_container: Iterable):
    """
    Returns whether the data container is a "groupable" container or not.

    By "groupable", we mean it is a 'categorical' or 'ordinal' variable.

    :param data_container: A generic container of data points.
    :type data_container: `iterable`
    """
    is_groupable = False
    if infer_data_type(data_container) in ["categorical", "ordinal"]:
        is_groupable = True
    return is_groupable


def num_discrete_groups(data_container: Iterable):
    """
    Returns the number of discrete groups present in a data container.

    :param data_container: A generic container of data points.
    :type data_container: `iterable`
    """
    return len(set(data_container))


def items_in_groups(data_container: Iterable):
    """
    Returns discrete groups present in a data container and the number items
    per group.

    :param data_container: A generic container of data points.
    :type data_container: `iterable`
    """
    return Counter(data_container)


def node_table(G, group_by=None, sort_by=None):
    """Return the node table of a graph G.

    ## Parameters

    - `G`: A NetworkX graph.
    - `group_by`: A key in the node attribute dictionary.
    - `sort_by`: A key in the node attribute dictionary.

    ## Returns

    A pandas DataFrame, such that the index is the node
    and the columns are node attributes.
    """
    node_table = []
    node_index = []
    for n, d in G.nodes(data=True):
        node_table.append(d)
        node_index.append(n)
    df = pd.DataFrame(data=node_table, index=node_index)
    df = group_and_sort(df, group_by, sort_by)
    return df


import networkx as nx


def edge_table(G) -> pd.DataFrame:
    """Return the edge table of a graph.

    The nodes involved in the edge are keyed
    under the `source` and `target` keys.
    This is a requirement for use with the hammer_bundle module
    in datashader's bundler.

    The rest of their node attributes are returned as columns.
    """
    data = []
    for u, v, d in G.edges(data=True):
        row = dict()
        row.update(d)
        row["source"] = u
        row["target"] = v
        data.append(row)
        if not G.is_directed():
            u, v = v, u
            row = dict()
            row.update(d)
            row["source"] = u
            row["target"] = v
            data.append(row)
    return pd.DataFrame(data)


from typing import Hashable, Iterable


def group_and_sort(
    node_table: pd.DataFrame, group_by: Hashable = None, sort_by: Hashable = None
) -> pd.DataFrame:
    """Group and sort a node table."""
    sort_criteria = []
    if group_by:
        sort_criteria.append(group_by)
    if sort_by:
        sort_criteria.append(sort_by)
    if sort_criteria:
        node_table = node_table.sort_values(sort_criteria)
    return node_table
